
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <map>

#include "Dispatcher.hh"


namespace Mf {


Notification::~Notification() {}


struct Dispatcher::Impl
{
	Impl() :
		id(1) {}

	Dispatcher::Handler getNewHandler()
	{
		id += 2;
		return (Dispatcher::Handler)id;
	}

	typedef std::pair<Dispatcher::Handler,Dispatcher::Function>	Callback;
	typedef std::multimap<std::string,Callback>					CallbackLookup;
	typedef CallbackLookup::iterator							CallbackIter;

	typedef std::multimap<Dispatcher::Handler,std::string>		HandlerLookup;
	typedef HandlerLookup::iterator								HandlerIter;


	inline Handler addHandler(const std::string& message,
			const Function& callback, Handler id)
	{
		callbacks.insert(std::make_pair(message, std::make_pair(id, callback)));
		handlers.insert(std::make_pair(id, message));

		return id;
	}

	inline void removeHandler(Handler id)
	{
		std::pair<HandlerIter,HandlerIter> matching(handlers.equal_range(id));

		for (HandlerIter it = matching.first; it != matching.second; ++it)
		{
			CallbackIter first = callbacks.find((*it).second);
			CallbackIter last = callbacks.end();

			for (CallbackIter jt = first; jt != last; ++jt)
			{
				if (((*jt).second).first == id)
				{
					callbacks.erase(jt);
					break;
				}
			}
		}

		handlers.erase(id);
	}

	unsigned long	id;

	CallbackLookup	callbacks;
	HandlerLookup	handlers;
};


Dispatcher::Dispatcher() :
	impl_(new Dispatcher::Impl) {}

Dispatcher::~Dispatcher() {}


Dispatcher& Dispatcher::getInstance()
{
	static Dispatcher dispatcher;
	return dispatcher;
}


Dispatcher::Handler Dispatcher::addHandler(const std::string& message,
		const Function& callback)
{
	return addHandler(message, callback, impl_->getNewHandler());
}

Dispatcher::Handler Dispatcher::addHandler(const std::string& message,
		const Function& callback, Handler id)
{
	// pass through
	return impl_->addHandler(message, callback, id);
}


void Dispatcher::removeHandler(Handler id)
{
	// pass through
	return impl_->removeHandler(id);
}


void Dispatcher::dispatch(const std::string& message, const Notification* param)
{
	std::pair<Impl::CallbackIter,Impl::CallbackIter>
		callbacks(impl_->callbacks.equal_range(message));

	for (Impl::CallbackIter it = callbacks.first; it != callbacks.second; ++it)
	{
		Function callback = ((*it).second).second;
		callback(param);
	}
}


} // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

